# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from cerebro.objects import MetaObject
from typing import Union, List


class GeneralTree(MetaObject):
    """
    This is base node to build general tree. The tree node can have children nodes.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """

    @property
    def parent(self):
        """
        Get node parent.
        :return: node parent.
        """
        return self._parent

    @property
    def is_root(self):
        """
        Check if this is root.
        :return: is root.
        """
        return self._parent is None

    @property
    def root(self):
        """
        Get tree root.
        :return: root node.
        """
        node = self
        while not node.is_root:
            node = node.parent
        return node

    @property
    def is_leaf(self):
        """
        Check if this is a leaf.
        :return: is leaf.
        """
        return len(self._nodes) == 0

    @property
    def nodes(self):
        """
        Get children nodes.
        :return: children nodes.
        """
        return [] if self.is_leaf else [child for child in self._nodes]

    def __init__(self, nodes=None, **kwargs):
        """
        Create new object.
        "param children: children nodes.
        :param kwargs:   keyword arguments.
        """
        super(GeneralTree, self).__init__(**kwargs)
        # Initialize attributes.
        self._parent = None
        # Attach children nodes.
        self._nodes = list()
        self.attach(nodes)

    def __data__(self, **kwargs) -> dict:
        """
        Get object data.
        :param kwargs:  keyword arguments.
        :return:        object data.
        """
        data = super(GeneralTree, self).__data__(**kwargs)
        if not self.is_leaf:
            data.update({'nodes': [node.__data__(**kwargs) for node in self._nodes]})
        return data

    def attach(self, nodes=None, **kwargs):
        """
        Attach children node(s) to tree.
        :param nodes:   node(s) to be attached.
        :param kwargs:  additional keyword arguments.
        :return:        node(s).
        """
        if nodes is None:
            return None
        # Attach single node.
        elif isinstance(nodes, GeneralTree):
            self.lock.acquire()
            self._nodes.append(nodes)
            nodes._parent = self
            self.lock.release()
            return nodes
        # Attach list of nodes.
        elif isinstance(nodes, list):
            self.lock.acquire()
            result = [self.attach(node, **kwargs) for node in nodes]
            self.lock.release()
            return result
        # Otherwise raise error because of invalid nodes.
        raise TypeError("Tree can only attach tree node(s)")

    def detach(self, indexes: Union[None, int, List[int]] = None, **kwargs):
        """
        Detach node(s) from tree.
        :param indexes: index(es) of node(s) to be detached.
        :param kwargs:  additional keyword arguments.
        :return:        node(s).
        """
        if indexes is None:
            return self, None
        # Detach single node.
        elif isinstance(indexes, int):
            self.lock.acquire()
            child = self._nodes.pop(indexes)
            child._parent = None
            self.lock.release()
            return child
        # Detach list of indexes.
        elif isinstance(indexes, list):
            self.lock.acquire()
            result = [self.detach(index, **kwargs) for index in indexes]
            self.lock.release()
            return result
        # Otherwise raise error because of invalid indexes.
        raise TypeError("Tree can only detach node(s) base on index(es) of them")

    def clean(self, **kwargs):
        """
        Clean all children.
        :param kwargs:  keyword arguments.
        """
        return self.detach([index for index in range(len(self._nodes))], **kwargs)

    def step(self, steps: Union[None, int, List[int]] = 0, **kwargs):
        """
        Step to another node from current node.
        :param steps:   steps to go.
        :param kwargs:  keyword arguments.
        :return:        destination node.
        """
        if steps is None:
            return self
        # Go single step.
        elif isinstance(steps, int):
            current = self
            # Go forward.
            if steps >= 0:
                current = self._nodes[steps]
            # Go backward.
            elif steps < 0:
                for _ in range(abs(steps)):
                    current = current.parent
            # Return result
            return current
        # Go list of steps.
        elif isinstance(steps, list):
            current = self
            for step in steps:
                current = current.step(step, **kwargs)
            return current
        # Otherwise raise error because of invalid steps.
        raise TypeError("Can only move on tree based on integer step(s)")



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

import os
import io
import sys
import base64
import logging
from cerebro.io import Path
from cerebro.designs import Singleton
from .records import VisualRecord
import matplotlib.image as mpimg
from matplotlib import pyplot as plt
from PIL import Image



# Store supported log levels.
LOG_LEVELS = {'none': logging.NOTSET, 'info': logging.INFO, 'debug': logging.DEBUG, 'warning': logging.WARNING,
              'error': logging.ERROR, 'critical': logging.CRITICAL}

# Default log format.
FORMAT = '%(asctime)s: %(levelname)s: %(message)s'

# Define logger name.
DEFAULT_LOGGER = 'cerebro_logger'
VISUAL_LOGGER = 'cerebro_visual_logger'


class Logging(metaclass=Singleton):
    """

    """

    @staticmethod
    def get_env_level():
        """
        Get log level from os environment.
        :return:    log level.
        """
        level = os.environ.get('DEBUG', 'none').lower()
        return LOG_LEVELS[level] if level in LOG_LEVELS else LOG_LEVELS['none']

    def __init__(self,
                 level=None,
                 log_dir: str = os.path.join(os.getcwd(), 'logs'),
                 log_file: str = 'log.txt',
                 fmt: str = FORMAT,
                 **kwargs):
        """
        Create new object.
        :param level:       log level.
        :param log_dir:     log directory.
        :param log_file:    log file.
        :param fmt:         log format.
        """
        super(Logging, self).__init__()
        self._filepath = None
        self._level = None
        self.setup(level, log_dir, log_file, fmt, **kwargs)

    def setup(self,
              level=None,
              log_dir: str = os.path.join(os.getcwd(), 'logs'),
              log_file: str = 'log.txt',
              fmt: str = FORMAT,
              **kwargs):
        """
        Setup the logging.
        :param level:       log level.
        :param log_dir:     log directory.
        :param log_file:    log file.
        :param fmt:         log format.
        :param kwargs:      keyword arguments.
        :return:            none.
        """
        level = os.environ.get('DEBUG', 'none').lower() if level is None else level
        level = LOG_LEVELS[level] if level in LOG_LEVELS else LOG_LEVELS['none']
        self._level = level
        # Initialize
        formatter = logging.Formatter(fmt)
        self._filepath = Path(os.path.join(log_dir, log_file))
        # Initialize logger.
        logger = logging.getLogger(DEFAULT_LOGGER)
        logger.setLevel(level)
        logger.handlers.clear()
        # Add console handler.
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
        # Add file handler.
        file_handler = logging.FileHandler(str(self._filepath.mkdir()), delay=True)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        # Initialize visual logger.
        visual_logger = logging.getLogger(VISUAL_LOGGER)
        visual_logger.setLevel(level)
        visual_logger.handlers.clear()
        # Add visual handler
        visual_handler = logging.FileHandler(str(self._filepath.mkdir().with_suffix('.html')), delay=True)
        visual_logger.addHandler(visual_handler)
        # Return result.
        return self

    @staticmethod
    def make_visual_log(title="", images=None, content="", fmt="jpg"):
        """
        Create a visual record for visual logging.
        :param title:   title of record.
        :param images:  images of record.
        :param content: content of record.
        :param fmt:     image format.
        :return:        visual record.
        """
        return VisualRecord(title, images, content, fmt)

    def info(self, msg):
        """
        Write info log.
        :param msg: message.
        :return:    logging.
        """
        if isinstance(msg, VisualRecord):
            logging.getLogger(VISUAL_LOGGER).info(str(msg))
            if self._level == logging.INFO:
                for image in msg.images:
                    plt.imshow(image)
                    plt.show()
        else:
            logging.getLogger(DEFAULT_LOGGER).info(str(msg))
        return self

    def debug(self, msg):
        """
        Write debug log.
        :param msg: message.
        :return:    logging.
        """
        if isinstance(msg, VisualRecord):
            logging.getLogger(VISUAL_LOGGER).debug(str(msg))
        else:
            logging.getLogger(DEFAULT_LOGGER).debug(str(msg))
        return self

    def warning(self, msg):
        """
        Write warning log.
        :param msg: message.
        :return:    logging.
        """
        if isinstance(msg, VisualRecord):
            logging.getLogger(VISUAL_LOGGER).warning(str(msg))
        else:
            logging.getLogger(DEFAULT_LOGGER).warning(str(msg))
        return self

    def error(self, msg):
        """
        Write error log.
        :param msg: message.
        :return:    logging.
        """
        if isinstance(msg, VisualRecord):
            logging.getLogger(VISUAL_LOGGER).error(str(msg))
        else:
            logging.getLogger(DEFAULT_LOGGER).error(str(msg))
        return self

    def critical(self, msg):
        """
        Write critical log.
        :param msg: message.
        :return:    logging.
        """
        if isinstance(msg, VisualRecord):
            logging.getLogger(VISUAL_LOGGER).critical(str(msg))
        else:
            logging.getLogger(DEFAULT_LOGGER).critical(str(msg))
        return self



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

import cv2
import base64
import numpy as np
from string import Template
from io import BytesIO as StringIO


def render_opencv(image, fmt="jpg"):
    """
    Render OpenCV image to base64.
    :param image:   given image.
    :param fmt:     image format.
    :return:        rendered image.
    """
    if not isinstance(image, np.ndarray):
        return None
    if image.shape[0] > 800:
        image = cv2.resize(image, [800, int(800 * image.shape[1] / image.shape[0])])
    if image.shape[1] > 800:
        image = cv2.resize(image, [int(800 * image.shape[0] / image.shape[1]), 800])
    val, buf = cv2.imencode(".%s" % fmt, image)
    return None if not val else buf, "image/%s" % fmt


def render_pil(image, fmt="jpg"):
    """
    Render PIL Image to base64.
    :param image:   given image.
    :param fmt:     image format.
    :return:        rendered image.
    """
    if not callable(getattr(image, "save", None)):
        return None
    output = StringIO()
    image.save(output, format=fmt)
    contents = output.getvalue()
    output.close()
    return contents, "image/%s" % fmt


def render_pylab(figure, fmt="jpg"):
    """
    Render figure to base64 image.
    :param figure:  given figure.
    :param fmt:     image format.
    :return:        rendered image.
    """
    if not callable(getattr(figure, "savefig", None)):
        return None
    output = StringIO()
    figure.savefig(output, format=fmt)
    contents = output.getvalue()
    output.close()
    return contents, "image/%s" % fmt


RENDERERS = [render_opencv, render_pil, render_pylab]


class VisualRecord(object):
    """
    This class is used to write visual log record.
    ---------
    @author:    Hieu Pham.
    @created:   20.10.2021.
    @updated:   20.10.2021.
    """

    def __init__(self, title="", images=None, content="", fmt="jpg"):
        """
        Create new object
        :param title:   title of log.
        :param images:  image of log.
        :param content: content of log.
        :param fmt:     image format.
        """
        self.title = title
        self.fmt = fmt
        # Assign image.
        if images is None:
            images = []
        self.images = images
        # Validate images.
        if not isinstance(images, (list, tuple, set, frozenset)):
            self.images = [self.images]
        # Assign footnotes.
        self.footnotes = content

    def __str__(self):
        """
        Get presentation string of object.
        :return:    string
        """
        t = Template(
            """
            <h4>$title</h4>
            $imgs
            $footnotes
            <hr/>""")
        # Return result.
        return t.substitute({
            "title": self.title,
            "imgs": self.render_images(),
            "footnotes": self.render_footnotes()
        })

    def render_images(self, no_html=False):
        """
        Render given images.
        :param no_html: no render html.
        :return:        rendered images.
        """
        rendered = []
        # Render images.
        for image in self.images:
            for renderer in RENDERERS:
                # Trying renderers we have one by one
                res = renderer(image, self.fmt)
                if res is None:
                    continue
                else:
                    rendered.append(res)
                    break
        # Return result.
        return "".join(
            Template('<img src="data:$mime;base64,$data" />').substitute({
                "data": base64.b64encode(data).decode(),
                "mime": mime
            }) for data, mime in rendered) if not no_html else [(np.squeeze(data), mine) for data, mine in rendered]

    def render_footnotes(self):
        """
        Render footnotes.
        :return:    rendered content.
        """
        if not self.footnotes:
            return ""
        # Return result.
        return Template("<pre>$footnotes</pre>").substitute({
            "footnotes": self.footnotes
        })



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

import json
from cerebro.objects import MetaObject as MetaObject
from cerebro.serializations import Serializer, Deserializer
from json import JSONEncoder as BaseJsonEncoder, JSONDecoder as BaseJsonDecoder


class JsonSerializer(BaseJsonEncoder, Serializer):
    """
    This class serialize object into json data.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """
    def __init__(self, **kwargs):
        """
        Create new object.
        :param kwargs:  keyword arguments.
        """
        Serializer.__init__(self, name='JsonEncoder')
        BaseJsonEncoder.__init__(self, default=self.__data__)

    def to_json(self, obj: MetaObject = None, indent: int = 4, sort: bool = False, **kwargs) -> str:
        """
        Encode object instance to json string.
        :param obj:     object to be encoded.
        :param indent:  intent level.
        :param sort:    sort keys or not.
        :param kwargs:  additional keyword arguments.
        :return:        json string.
        """
        return json.dumps(self.data(obj), indent=indent, sort_keys=sort, **kwargs)

    def to_file(self, obj: MetaObject = None, path: str = "default.json", intent: int = 4, sort: bool = False, **kwargs):
        """
        Encode object instance to json file.
        :param obj:     object to be encoded.
        :param path:    path to json file.
        :param intent:  intent level.
        :param sort:    sort keys or not.
        :param kwargs:  additional keyword arguments.
        :return:        none.
        """
        with open(path, 'w+') as file:
            file.seek(0)
            file.write(self.to_json(obj, intent, sort, **kwargs))


class JsonDeserializer(BaseJsonDecoder, Deserializer):
    """
    This class deserialize json data to object.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """
    def __init__(self, **kwargs):
        """
        Create new object.
        :param kwargs:  keyword arguments.
        """
        Deserializer.__init__(self, name='JsonDecoder')
        BaseJsonDecoder.__init__(self, object_hook=self.create)

    def from_json(self, data: str = None, **kwargs):
        """
        Decode object from json string.
        :param data:    json data.
        :param kwargs:  keyword arguments.
        :return:        decoded object.
        """
        return self.decode(data) if data is not None else None

    def from_file(self, path: str = "default.json", **kwargs):
        """
        Decode object from json file.
        :param path:    path to json file.
        :param kwargs:  additional keyword arguments.
        :return:        decoded object.
        """
        with open(path, "r+") as file:
            file.seek(0)
            return self.from_json(file.read())



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from abc import ABC, abstractmethod
from typing import Any, Union, List, Callable


class BatchProcess(ABC):
    """
    This pattern is used when we need to process batch of things but every
    item will be processed in the same way.
    ---------
    @author:    Hieu Pham.
    @created:   15.10.2021.
    @updated:   20.10.2021.
    """

    def batch_process(self, payload: Union[Any, List[Any]] = None, hook: Callable = None, **kwargs):
        """
        Process batch of payloads.
        :param payload: batch of payloads.
        :param hook:    hook to check payload.
        :param kwargs:  keyword arguments.
        :return:        any.
        """
        if payload is None:
            return None
        # In case of batch is list, process each item sequentially.
        if isinstance(payload, list) if hook is None else hook(payload):
            result = [self.batch_process(single, **kwargs) for single in payload]
        # In case of batch is not list, process it as singular object.
        else:
            result = self.single_process(payload, **kwargs)
        # Return result.
        return result

    @abstractmethod
    def single_process(self, payload: Any = None, **kwargs):
        """
        Process singular object.
        :param payload: given object.
        :param kwargs:  keyword arguments.
        :return:        any.
        """
        return payload



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from abc import ABCMeta
from threading import RLock


class MetaClass(ABCMeta):
    """
    The meta class will be used to control python object at low level.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """
    pass


class MetaObject(metaclass=MetaClass):
    """
    The meta object is low level object of all inherited objects.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """

    @property
    def name(self):
        """
        Get object name.
        :return: object name.
        """
        return self._name

    @name.setter
    def name(self, value: str = None):
        """
        Set object name.
        :param value: given name.
        :return:      none.
        """
        self._name = value

    @property
    def data(self):
        """
        Get object data.
        :return: object data.
        """
        return self.__data__()

    @property
    def lock(self):
        """
        Get thread locking.
        :return: thread locking.
        """
        return self._lock

    def __init__(self, name: str = None, **kwargs):
        """
        Create new object.
        :param name:    object name.
        :param kwargs:  keyword arguments.
        """
        self._name = name
        self._lock = RLock()
        super(MetaObject, self).__init__()

    def __data__(self, **kwargs) -> dict:
        """
        Generate object data.
        :param kwargs:  keyword arguments.
        :return:        object data.
        """
        data = {'classname': self.__class__.__name__, 'module': self.__module__}
        if self._name is not None:
            data.update({'name': self._name})
        return data



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

import os
import shutil
from copy import copy
from datetime import datetime
from pathlib import Path as BasePath


class Path:
    """
    This class is used to handle a path.
    ---------
    @author:    Hieu Pham.
    @created:   19.10.2021.
    @updated:   20.10.2021.
    """

    def __init__(self, path: str = None, **kwargs):
        """
        Created new object.
        :param path:    given path.
        :param kwargs:  keyword arguments.
        """
        super().__init__()
        self._path = BasePath(os.path.expanduser(path))

    def __str__(self):
        """
        Get string represent of path.
        :return:    path string.
        """
        return str(self._path)

    def insert(self, index, candidate, **kwargs):
        """
        Insert a candidate into current path.
        :param index:       index to insert.
        :param candidate:   candidate to insert.
        :param kwargs:      keyword arguments.
        :return:            path object.
        """
        parts = list(self._path.parts)
        parts.insert(index, candidate)
        self._path = BasePath(*parts)
        return self

    def remove(self, index, **kwargs):
        """
        Remove a candidate from current path.
        :param index:   index to be removed.
        :param kwargs:  keyword arguments.
        :return:        path object.
        """
        parts = list(self._path.parts)
        parts.pop(index)
        self._path = BasePath(*parts)
        return self

    def change(self, index, candidate, **kwargs):
        """
        Change a part of path with a candidate.
        :param index:       index to be replaced.
        :param candidate:   candidate to replace.
        :param kwargs:      keyword arguments.
        :return:            path object.
        """
        parts = list(self._path.parts)
        parts[index] = candidate
        self._path = BasePath(*parts)
        return self

    def with_suffix(self, suffix: str = None):
        """
        Replace suffix with another.
        :param suffix:  desired suffix.
        :return:        path object.
        """
        if suffix is not None:
            self._path = self._path.with_suffix(suffix)
        return self

    def unique(self, **kwargs):
        """
        Insert unique number to suffix to prevent duplicate.
        :param kwargs:  keyword arguments.
        :return:        path object.
        """
        count = 0
        while True:
            stem = '%s-%s' % (self._path.stem, count) if count > 0 else self._path.stem
            new_path = copy(self._path)
            new_path.with_stem(stem)
            if new_path.is_dir() or new_path.is_file():
                count += 1
                continue
            self._path = new_path
            return self

    def datetime(self, formats: str = '-%Y-%m-%d-%H-%M-%S', **kwargs):
        """
        Insert datetime to suffix to prevent duplicate.
        :param formats: string format of time.
        :param kwargs:  keyword arguments.
        :return:        path object.
        """
        stem = '%s-%s' % (self._path.stem, datetime.now().strftime(formats))
        self._path = self._path.with_stem(stem)
        return self

    def mkdir(self, **kwargs):
        """
        Make sure the directory is exist.
        :param kwargs:  additional keyword arguments.
        :return:        none.
        """
        if len(self._path.suffixes) == 0:
            os.makedirs(self._path, exist_ok=True)
        else:
            os.makedirs(self._path.parent, exist_ok=True)
        return self

    def copy(self, dst=None, **kwargs):
        """
        Copy file or directory of this path to a desired destination.
        :param dst:     desired destination.
        :param kwargs:  additional keyword arguments.
        :return:        none.
        """
        assert isinstance(dst, Path), 'Destination must be a path object.'
        # In case both of source and destination is directory.
        if len(self._path.suffixes) == 0 and len(dst._path.suffixes) == 0:
            shutil.copytree(str(self), str(dst))
        # In case both of source and destination is file.
        if len(self._path.suffixes) != 0 and len(dst._path.suffixes) != 0:
            shutil.copy2(str(self), str(dst))
        # Otherwise, raise error.
        raise IOError('Source and destination are not same type.')

    def move(self, dst=None, **kwargs):
        """
        Move file or directory of this path to a desired destination.
        :param dst:     desired destination.
        :param kwargs:  additional keyword arguments.
        :return:        none.
        """
        assert isinstance(dst, Path), 'Destination must be a path object.'
        shutil.move(str(self), str(dst))

    def rm(self, **kwargs):
        """
        Remove file or directory of this path.
        :param kwargs:  additional keyword arguments.
        :return:        none.
        """
        if self._path.is_file():
            os.remove(str(self))
        else:
            shutil.rmtree(str(self))



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from copy import copy
from typing import Union, Dict
from cerebro.datastructs import GeneralTree


class Component(GeneralTree):
    """
    Compose objects into tree structures and then work with these structures
    as if they were individual objects.
    ---------
    @author:    Hieu Pham.
    @created:   15.10.2021.
    @updated:   20.10.2021.
    """

    def __init__(self, prior: int = 1, sociable: bool = True, **kwargs):
        """
        Create new instance.
        :param prior:       priority of process.
        :param sociable:    update result from siblings.
        :param kwargs:      additional keyword arguments.
        """
        self.prior = prior
        self.sociable = sociable
        super().__init__(**kwargs)

    def __call__(self, *args, **kwargs) -> Union[None, dict]:
        """
        Override call function treats object instance as callable function.
        :param args:    arguments to be passed.
        :param kwargs:  keyword arguments to be passed.
        :return:        result data.
        """
        inputs = copy(kwargs)
        # In case of priority is negative, process as preorder.
        if self.prior < 0:
            outputs = self.process(**kwargs)
            if isinstance(outputs, dict):
                kwargs.update(outputs)
        # Process all children components.
        for child in self.nodes:
            outputs = child(**kwargs) if child.sociable else child(**inputs)
            if isinstance(outputs, dict):
                kwargs.update(outputs)
        # In case of priority is positive, process as postorder.
        if self.prior > 0:
            outputs = self.process(**kwargs)
            if isinstance(outputs, dict):
                kwargs.update(outputs)
        # Return result.
        return kwargs

    def process(self, **kwargs) -> Union[None, Dict]:
        """
        Applies processing steps.
        :param kwargs:  keyword arguments.
        :return:        result data.
        """
        return dict()

    def __data__(self, **kwargs) -> dict:
        """
        Get object data.
        :param kwargs:  keyword arguments.
        :return:        object data.
        """
        data = {'prior': self.prior, 'sociable': self.sociable}
        data.update(super(Component, self).__data__(**kwargs))
        return data

    def attach(self, nodes=None, **kwargs):
        """
        Attach other components.
        :param nodes:   children component(s) to be attached.
        :param kwargs:  additional keyword arguments to be passed.
        :return:        children component(s).
        """
        # Attach single component.
        if nodes is None or isinstance(nodes, Component):
            return super().attach(nodes, **kwargs)
        # Attach list of components.
        elif isinstance(nodes, list):
            return [self.attach(child, **kwargs) for child in nodes]
        # Raise error if children is not valid.
        raise TypeError('Component can only attach component(s).')



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from cerebro.objects import MetaClass
from threading import RLock


class Singleton(MetaClass):
    """
    Singleton pattern ensure that there is only one instance will be created.
    ---------
    @author:    Hieu Pham.
    @created:   25.09.2021.
    @updated:   20.10.2021.
    """

    # We will keep created instance.
    _instances = {}
    # Locking for thread safe.
    _lock = RLock()

    # We now have a lock object that will be used to synchronize threads during
    # first access to the Singleton.
    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect the returned instance.
        :param args:    additional arguments.
        :param kwargs:  additional keyword arguments.
        """
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.
        cls._lock.acquire()
        # The first thread to acquire the lock, reaches this conditional,
        # goes inside and creates the Singleton instance. Once it leaves the
        # lock block, a thread that might have been waiting for the lock
        # release may then enter this section. But since the Singleton field
        # is already initialized, the thread won't create a new object.
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        cls._lock.release()
        # Return result.
        return cls._instances[cls]



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

import sys
from distutils.core import setup
from setuptools import find_packages


def get_readme_description(src: str = 'README.md'):
    """
    Get readme description.
    :param src: readme file.
    :return:    description.
    """
    if sys.version_info[0] < 3:
        with open(src) as f:
            readme = f.read()
    else:
        with open(src, encoding='utf-8') as f:
            readme = f.read()
    return readme


setup(
    name='cerebro-toolkit',
    packages=find_packages(),
    namespace_packages=['cerebro'],
    version='0.0.1',
    license='MIT',
    zip_safe=True,
    description='Provide systematic process of build and improving codebase without creating new functionality that '
                'can transform a mess into clean code and simple design architect.',
    long_description=get_readme_description('README.md'),
    long_description_content_type='text/markdown',
    author='Hieu Pham',
    author_email='hieupt.ai@gmail.com',
    url='https://github.com/hieupth/cerebro-toolkit',
    download_url='https://github.com/hieupth/cerebro-toolkit/archive/v_01.tar.gz',
    keywords=['refactoring', 'data structures', 'design patterns'],
    install_requires=[],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
)


# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from importlib import import_module
from cerebro.designs import Singleton
from cerebro.objects import MetaObject


class Serializer(MetaObject, metaclass=Singleton):
    """
    Encoder can serialize objects into object data. The encoder will be a singleton
    to reduce memory consuming.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """
    def __data__(self, instance: MetaObject = None, **kwargs) -> dict:
        """
        Get object data.
        :param instance: instance to be encoded.
        :param kwargs:   keyword arguments.
        :return:         object data.
        """
        return instance.data(**kwargs) if isinstance(instance, MetaObject) else None


class Deserializer(MetaObject, metaclass=Singleton):
    """
    Decoder can deserialize object data into objects. The decoder will be a singleton
    to reduce memory consuming.
    ---------
    @author:    Hieu Pham.
    @created:   10.10.2021.
    @updated:   20.10.2021.
    """
    @classmethod
    def create(cls, data: dict = None, **kwargs):
        """
        Create object from data.
        :param data:    object data.
        :param kwargs:  keyword arguments.
        :return:        object.
        """
        # Validate data format. The correct format must have classname and module info.
        if 'classname' in data and 'module' in data:
            # Dynamic load module and class based on classname.
            classname, module = data.pop('classname'), data.pop('module')
            classes = getattr(import_module(module), classname)
            # Deserialize object if class is serializable.
            if issubclass(classes, MetaObject):
                return classes(**data)
        # Otherwise, raise an error.
        raise AssertionError('Object data format is invalid!')



# ------------------------------------------------------------------------------
#  MIT License
#
#  Copyright (c) 2021 Hieu Pham. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
# ------------------------------------------------------------------------------

from typing import Any
from cerebro.objects import MetaObject


class Stack(MetaObject):
    """
    Stack is an implementation of first in last out structure.
    ---------
    @author:    Hieu Pham.
    @created:   13.10.2021.
    @updated:   20.10.2021.
    """

    @property
    def size(self):
        """
        Get stack size.
        :return: size.
        """
        return len(self._stack)

    def __init__(self, **kwargs):
        """
        Create new object.
        :param kwargs:  keyword arguments.
        """
        self._stack = list()
        super(Stack, self).__init__(**kwargs)

    def push(self, obj: Any = None, **kwargs):
        """
        Push object into stack.
        :param obj:     object to be pushed.
        :param kwargs:  keyword arguments.
        :return:        object.
        """
        if obj is not None:
            self.lock.acquire()
            self._stack.append(obj)
            self.lock.release()
            return obj

    def pop(self, **kwargs):
        """
        Pop object from stack.
        :param kwargs:  keyword arguments.
        :return:        object.
        """
        self.lock.acquire()
        result = self._stack.pop()
        self.lock.release()
        return result

    def get(self, index: int = 0):
        """
        Get item based on index.
        :param index:   index.
        :return:        item.
        """
        return self._stack[index]

    def clean(self, **kwargs):
        """
        Pop all objects from stack.
        :param kwargs:  keyword arguments.
        :return:        object.
        """
        self.lock.acquire()
        result = [self._stack.pop() for i in range(len(self._stack))]
        self.lock.release()
        return result
